import argparse
from itertools import islice

import wordgame

_GAME_CHOICES = [
    'countdown',
]

_DEFAULT_GAME_CHOICE = _GAME_CHOICES[0]


def _make_parser():
    parser = argparse.ArgumentParser(
        prog='wordgame',
        description='Solve for wordgames such as scrabble and countdown.',
    )
    parser.add_argument(
        'letters',
        metavar='LETTERS',
        help='the letters to use',
    )
    parser.add_argument(
        '-g', '--game',
        metavar='GAME',
        choices=_GAME_CHOICES,
        default=_DEFAULT_GAME_CHOICE,
        help='which wordgame to play',
    )
    parser.add_argument(
        '-n', '--maximum-number-of-results',
        type=int,
        help='maximum number of results to return (default: all results)',
    )
    return parser


def take(n, sequence):
    try:
        return islice(sequence, n)
    except ValueError:
        return []


def _main():
    parser = _make_parser()
    settings = parser.parse_args()
    if settings.game != 'countdown':
        raise NotImplementedError(
            'Only the countdown game type is currently supported.')
    results = wordgame.countdown(settings.letters)

    results = take(
        settings.maximum_number_of_results,
        results)

    for result in results:
        print(result)


if __name__ == '__main__':
    _main()



import unittest

from wordgame.__main__ import _make_parser


class TestMakeParser(unittest.TestCase):
    def setUp(self):
        self.parser = _make_parser()

    def test_game_argument_default(self):
        settings = self.parser.parse_args(['abc'])
        self.assertEqual(settings.game, 'countdown')
        self.assertIsNone(settings.maximum_number_of_results)

    def test_number(self):
        settings = self.parser.parse_args(['abc', '-n', '5'])
        self.assertEqual(settings.maximum_number_of_results, 5)



import unittest

from wordgame.__main__ import take


class TestTake(unittest.TestCase):
    def test_number_of_None(self):
        expected = [0, 1, 2]
        actual = list(take(None, range(3)))
        self.assertEqual(actual, expected)

    def test_number_of_zero(self):
        expected = []
        actual = list(take(0, range(3)))
        self.assertEqual(actual, expected)

    def test_number_same(self):
        expected = [0, 1, 2]
        actual = list(take(3, range(3)))
        self.assertEqual(actual, expected)

    def test_number_smaller(self):
        expected = [0, 1]
        actual = list(take(2, range(3)))
        self.assertEqual(actual, expected)

    def test_number_larger(self):
        expected = [0, 1, 2]
        actual = list(take(5, range(3)))
        self.assertEqual(actual, expected)

    def test_number_negative(self):
        expected = []
        actual = list(take(-1, range(3)))
        self.assertEqual(actual, expected)



import unittest

from wordgame import sort_words


class TestSortWords(unittest.TestCase):
    def test_empty(self):
        expected = []
        actual = list(sort_words([]))
        self.assertEqual(expected, actual)

    def test_singular(self):
        expected = ['a']
        actual = list(sort_words(['a']))
        self.assertEqual(expected, actual)

    def test_prefix(self):
        expected = [
            'leisurely',
            'aardvark',
            'idea',
            'word',
            'bad',
            'i',
        ]
        actual = list(sort_words([
            'i',
            'idea',
            'word',
            'aardvark',
            'bad',
            'leisurely',
        ]))
        self.assertEqual(expected, actual)



import unittest

from wordgame import _find_list_of_words


class TestWordsList(unittest.TestCase):
    '''Tests that the contents of the words.txt file are valid'''
    @classmethod
    def setUpClass(cls):
        cls.words = list(_find_list_of_words())

    def test_no_empty_string_words(self):
        for word in self.words:
            self.assertNotEqual(len(word), 0)

    def test_no_whitespace_are_present(self):
        for word in self.words:
            for character in word:
                self.assertFalse(character.isspace())

    def test_all_words_are_alphabetical(self):
        for word in self.words:
            self.assertTrue(word.isalpha())



from collections import Counter

from string import ascii_letters

_COUNTDOWN_LETTER_COUNT = 9
_COUNTDOWN_LETTER_COUNT_FOR_TEASER = 8


def countdown(letters, words=None, sort=True, skip_letters_validation=False):
    if not skip_letters_validation:
        letters = validate_countdown_letters(letters)
    else:
        letters = letters.lower()

    if words is None:
        words = _find_list_of_words()
    words = (word.lower() for word in words)

    allowed_words_by_length = filter(
        length_predicate(max_length=len(letters)),
        words
    )
    results = (
        word
        for word in allowed_words_by_length
        if _can_word_be_made_with_letters(word=word, letters=letters))

    if sort:
        results = sort_words(results)

    return results


# From: http://www.thecountdownpage.com/letters.htm
_LETTER_DISTRIBUTION = {
    'a': 15,
    'e': 21,
    'i': 13,
    'o': 13,
    'u': 5,

    'b': 2,
    'c': 3,
    'd': 6,
    'f': 2,
    'g': 3,
    'h': 2,
    'j': 1,
    'k': 1,
    'l': 5,
    'm': 4,
    'n': 8,
    'p': 4,
    'q': 1,
    'r': 9,
    's': 9,
    't': 9,
    'v': 1,
    'w': 1,
    'x': 1,
    'y': 1,
    'z': 1,
}


def validate_countdown_letters(letters):
    if len(letters) not in [
            _COUNTDOWN_LETTER_COUNT,
            _COUNTDOWN_LETTER_COUNT_FOR_TEASER]:
        raise ValueError(
            f'There must be {_COUNTDOWN_LETTER_COUNT} letters '
            f'or {_COUNTDOWN_LETTER_COUNT_FOR_TEASER} for teasers '
            'to play countdown.')
    for letter in letters:
        if letter not in ascii_letters:
            raise ValueError(
                f'{repr(letter)} is not a valid letter in countdown')
    letters = letters.lower()
    for letter in sorted(list(set(letters))):
        count_of_a_letter = letters.count(letter)
        max_allowed = _LETTER_DISTRIBUTION[letter]
        if max_allowed < count_of_a_letter:
            raise ValueError(
                f'There are too many {letter}s found in the countdown letters '
                f'{count_of_a_letter} found '
                f'but only {max_allowed} are allowed')
    return letters


class PartiallyOrderedCounter(Counter):
    '''Source: https://stackoverflow.com/a/44644868/1011785'''

    def __le__(self, other):
        """ Multiset inclusion """
        return all(v <= other[k] for k, v in self.items())

    def __lt__(self, other):
        """ Multiset strict inclusion """
        return self <= other and self != other

    # TODO : __ge__ and __gt__
    # Beware : they CANNOT be written in terms of __le__ or __lt__


def _can_word_be_made_with_letters(letters, word):
    return PartiallyOrderedCounter(word) <= PartiallyOrderedCounter(letters)


def _find_list_of_words():
    with open('words.txt', 'r') as f:
        for line in f:
            yield line.strip()


def length_predicate(max_length):
    def pred(value):
        return len(value) <= max_length
    return pred


def sort_words(words):
    return sorted(
        sorted(words),
        key=len,
        reverse=True)



from setuptools import setup

setup(
    name='wordgame',
    version='0.1.0',
    author='Aaron Robson',
    author_email='arobsonse@hotmail.co.uk',
    packages=['wordgame', 'wordgame.test'],
    url='https://github.com/AaronRobson/wordgame/',
    license='LICENSE.txt',
    description='A solver for wordgames such as scrabble and countdown.',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    install_requires=[],
    include_package_data=True,
    python_requires='>=3.6',
)



import unittest
from unittest.mock import patch

import wordgame

_WORDS = [
    'bad',
    'rejoiced',
    'swearing',
]


class TestCountdown(unittest.TestCase):
    def setUp(self):
        self._WORDS_UPPER = [word.upper() for word in _WORDS]

    def test_simple(self):
        actual = list(wordgame.countdown(
            'abcd',
            words=_WORDS,
            skip_letters_validation=True))
        expected = 'bad'
        self.assertIn(expected, actual)

    def test_upper_case_letters(self):
        actual = list(wordgame.countdown(
            'ABCD',
            words=_WORDS,
            skip_letters_validation=True))
        expected = 'bad'
        self.assertIn(expected, actual)

    def test_upper_case_words(self):
        actual = list(wordgame.countdown(
            'abcd',
            words=self._WORDS_UPPER,
            skip_letters_validation=True))
        expected = 'bad'
        self.assertIn(expected, actual)

    def test_upper_case_letters_and_words(self):
        actual = list(wordgame.countdown(
            'ABCD',
            words=self._WORDS_UPPER,
            skip_letters_validation=True))
        expected = 'bad'
        self.assertIn(expected, actual)

    def test_swearing(self):
        '''featured in 8 out 10 cats does Countdown'''
        actual = list(wordgame.countdown(
            'aswinger',
            words=_WORDS,
            skip_letters_validation=True))
        expected = 'swearing'
        self.assertIn(expected, actual)

    def test_rejoiced(self):
        '''featured in 8 out 10 cats does Countdown'''
        actual = list(wordgame.countdown(
            'joecried',
            words=_WORDS,
            skip_letters_validation=True))
        expected = 'rejoiced'
        self.assertIn(expected, actual)


class TestCanWordBeMadeWithLetters(unittest.TestCase):
    def test_positive(self):
        actual = wordgame._can_word_be_made_with_letters(
            word='bad',
            letters='abcd')
        self.assertTrue(actual)

    def test_negative(self):
        actual = wordgame._can_word_be_made_with_letters(
            word='other',
            letters='abcd')
        self.assertFalse(actual)

    def test_swearing(self):
        '''featured in 8 out 10 cats does Countdown'''
        actual = wordgame._can_word_be_made_with_letters(
            word='swearing',
            letters='aswinger')
        self.assertTrue(actual)

    def test_rejoiced(self):
        '''featured in 8 out 10 cats does Countdown'''
        actual = wordgame._can_word_be_made_with_letters(
            word='rejoiced',
            letters='joecried')
        self.assertTrue(actual)


class TestValidateCountdownLetters(unittest.TestCase):
    def test_valid_letters(self):
        expected = given = 'fdofieasg'
        actual = wordgame.validate_countdown_letters(
            letters=given)
        self.assertEqual(actual, expected)

    def test_invalid_length(self):
        with self.assertRaises(ValueError):
            wordgame.validate_countdown_letters(
                letters='abc'
            )

    def test_invalid_characters(self):
        with self.assertRaises(ValueError):
            wordgame.validate_countdown_letters(
                letters='abcdefgh#'
            )

    def test_invalid_amount_of_a_letter(self):
        with self.assertRaises(ValueError):
            wordgame.validate_countdown_letters(
                letters='xxabcdefg'
            )


@patch("builtins.open")
class TestFindListOfWords(unittest.TestCase):
    def test_trailing_new_lines_are_removed(self, mock_open):
        given = ['abc\n', 'def\n']
        mock_open.return_value.__enter__ = mock_open
        mock_open.return_value.__iter__.return_value = iter(given)
        actual = list(wordgame._find_list_of_words())
        expected = ['abc', 'def']
        self.assertEqual(actual, expected)
